-- -- -- -- -- -- -- -- -- -- -- -- 
-- Star Trek Armada SOD Importer
-- SOD Versions: 1.70 - 1.93
-- Importer version 1.0.1
-- Date: Aug 30, 2013
-- Author: Mr. Vulcan (Taras Pankov)
-- TPankov (at) yahoo (dot) ca

-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
-- This Importer is for SOD Rendering only --
-- DO NOT USE TO CREATE MODELS -- ONLY FOR RENDERING --
-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
--  Use  responsibly!  Modding communities are built on trust and respect for fellow modders.  --
-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --

fn ImportSODRender FullFilePath:"" TexturePath:"" =
(
global vers = 0.0
global num_lighting = 0

include "SODImportSupport.ms"
-- Supporting Functions --
--------------------------
function GetSaveFile =
(
	-- Get file name for importing
	filePath = getOpenFileName caption:"Import SOD for RENDERING"  \
			filename:("$max\meshes\*.sod") \
			types:"Armada II Storm 3D Object|*.sod|"
	if (getfiles filePath).count == 0 then (filePath = undefined)

	if filePath == Undefined then
	(
		print "Cancel"
		f = Undefined
	)
	else (f = fopen filePath "rb")
) -- end GetSaveFile
--------------------

----------------------
-- Read functions --

function ReadHeader fID =
(
	str = ""
	for i1 = 1 to 10 do
	(
		str = str + (bit.IntAsChar (readByte fID))
	)
	vers = ReadFloat fID
)

-- Reset the materials and scene
function resetEnvironment =
(
	delete objects
	
	for i1 = 1 to meditmaterials.count do
	(
		meditMaterials[i1] = standardMaterial name:((i1 as string) + " - DefMat")
	)	
		
ret = true
)
---------------------------------------
-- ReadLightingMaterials
function ReadLightingMaterials fID =
(
	--try (
	include "SODImportSupport.ms"
	num_lighting = ReadShort fID #unsigned
	if (num_lighting > 24) then
	(
		messageBox "Warning: Model has more than 24 materials. Meshes with multiple lighting materials may not render properly."
	)
	
	--catch (
	--format " % ***\n" (getCurrentException())

	for i1 = 1 to (amin num_lighting 24) do
	(
		-- Material Name
		mID = ReadIdentifier fID
		meditMaterials[i1].name = mID + "_01"
		meditMaterials[i1].ShowInViewport = on
		meditMaterials[i1].adLock = false
		meditMaterials[i1].ambient = ReadColour fID
		meditMaterials[i1].diffuse = ReadColour fID
		specularColour = ReadColour fID
		meditMaterials[i1].Specular = color (specularColour.red*2) (specularColour.green*2) (specularColour.blue*2)
		if ((specularColour.red + specularColour.green + specularColour.blue) < 0.1) then (meditMaterials[i1].specularLevel = 0)
		else (meditMaterials[i1].specularLevel = 50)
		meditMaterials[i1].glossiness = ((ReadFloat fID)*2)
		matModel = ReadByte fID #unsigned -- not used in max

		-- 0/1 for alpha map illumination
		if (vers > 1.8001) then
		(
			illumParam = ReadByte fID #unsigned
			if illumParam > 0 then
				meditMaterials[i1].selfillumMapEnable = on
			else
				meditMaterials[i1].selfillumMapEnable = off
		)
	)
)-- end ReadLightingMaterials

-------------------------------
-- Node Definitions
function ReadNodes fID TexPath =
(
include "SODImportSupport.ms"

num_nodes = ReadShort fID #unsigned

for i1 = 1 to num_nodes do
(
node_type = ReadShort fID #unsigned
node_name = ReadIdentifier fID

-- add the command syntax for special nodes
cmd_code = ""
case of
(
	(node_type == 12): (cmd_code = "e_") --emitter
	(node_type == 3): (cmd_code = "s_")  --sprite
)
node_name = cmd_code + node_name

-- node parent
node_par_name = ReadIdentifier fID
node_par = nodeByName node_par_name

-- local transform between parent and current node
node_M = ReadTransform fID
node_M = XYZtoRUF node_M
-- global transform
if (node_par == undefined) then node_M = node_M
else (node_M = (node_M*node_par.transform))

-- Read type-specific info
-- emitter
if node_type == 12 then
(
	emName = ReadIdentifier fID -- unused in max
	if (node_par == undefined) then
	(
		Point transform:node_M name:node_name wirecolor:(color 87 224 87) axistripod:on cross:off -- green
	)
	else
	(
		Point transform:node_M name:node_name parent:node_par wirecolor:(color 87 224 87) axistripod:on cross:off -- green
	)
)

-- sprite
if node_type == 3 then
(
	if (node_par == undefined) then
	(
		Point transform:node_M name:node_name wirecolor:(color 88 199 225) axistripod:on cross:off -- blue
	)
	else
	(
		Point transform:node_M name:node_name parent:node_par wirecolor:(color 88 199 225) axistripod:on cross:off -- blue
	)
)

-- hardpoint / null
if node_type == 0 then
(
	if (node_par == undefined) then
	(
		Point transform:node_M name:node_name wirecolor:(color 134 6 6) axistripod:on cross:off -- dark red
	)
	else
	(
		Point transform:node_M name:node_name parent:node_par wirecolor:(color 134 6 6) axistripod:on cross:off -- dark red
	)
)

-- mesh
if node_type == 1 then
(--
	-- blending type
	node_texBld = ReadIdentifier fID
	
	-- bump map params and lighting params, use is uncertain
	bumpm = 0
	if (vers > 1.9101) then
	(
		dummy = ReadLong fID #unsigned -- unused
		bumpm = ReadLong fID #unsigned
	)
	
	-- texture file
	node_texFile = ReadIdentifier fID
	if (TexPath.count > 1) then
	(
		node_texFile = TexPath + "\\" + node_texFile
	)
	-- unknown parameters
	if (vers == 1.91) then
	(
		dummy = ReadShort fID #unsigned
	)
	node_borg = ""
	if (vers > 1.9101) then
	(
		dummy = ReadShort fID #unsigned
		dummy = ReadShort fID #unsigned
		
		-- Bump map parameters
		if bumpm == 2 then
		(
			node_bump = ReadIdentifier fID
			if (TexPath.count > 1) then
			(
				node_bump = TexPath + "\\" + node_bump
			)
			dummy = ReadShort fID #unsigned
			dummy = ReadShort fID #unsigned
		)
		
		-- Borg texture
		--if () then
		--(
		node_borg = ReadIdentifier fID
		if (TexPath.count > 1) then
		(
			node_borg = TexPath + "\\" + node_borg
		)
		dummy = ReadShort fID
		--)
		
	) -- vers > 1.91
	
	-- Number of vertices, texture vertices, and lighting groups
	node_verts = ReadShort fID
	node_tverts = ReadShort fID
	node_lgroups = ReadShort fID
	
	-- Read in the vertex array
	node_pos = node_M.row4
	vert_array = #()
	for i2 = 1 to node_verts do
	(
		verti = [(ReadFloat fID), (ReadFloat fID), (ReadFloat fID)]
		--append vert_array [(verti[1]+node_pos[1]), (verti[3]+node_pos[3]), (verti[2]+node_pos[2])]
		append vert_array [(verti[1]), (verti[3]), (verti[2])]

 	)
	
	-- Read in the texture vertex array
	tvert_array = #()
	for i2 = 1 to node_tverts do
	(
		append tvert_array [(ReadFloat fID), (1-(ReadFloat fID)), 0]
 	)

	-- Lighting groups array
	fface_array = #()
	tface_array = #()
	num_faces_array = #()
	group_lighting_mat = #()
	--group_faces = #()
	--lcnt = 0
	for i2 = 1 to node_lgroups do
	(
		num_faces_group = ReadShort fID
		append num_faces_array num_faces_group
		append group_lighting_mat (ReadIdentifier fID)
		-- make a list of faces for this group
		--group_faces[i1] = #{(lcnt+1)..(lcnt+num_faces_group)}
		--lcnt = lcnt+num_faces_group
		-- Now read in the actual faces in this group
		for i3 = 1 to num_faces_group do
		(
			fface = [1,1,1]
			tface = [1,1,1]
			for i4 = 1 to 3 do
			(
				fface[i4] = ((ReadShort fID #unsigned)+1)
				tface[i4] = ((ReadShort fID #unsigned)+1)
			)
			append fface_array fface
			append tface_array tface
		)
	)
	
	-- Now we can create the mesh
	mnode = mesh vertices:vert_array tverts:tvert_array faces:fface_array
	-- note: some versions of 3dsmax have broken texture face creation.
	-- 3dsMax recommends the following workaround:
	mmesh = mnode.mesh
	buildTVFaces mmesh
	for i = 1 to mmesh.numfaces do (setTVFace mmesh i tface_array[i])
	
	-- assign smoothing groups based on lighting groups
	lcnt = 1
	for i2 = 1 to node_lgroups do
	(
		for i3 = lcnt to (lcnt - 1 + num_faces_array[i2]) do
		(
			setFaceSmoothGroup mnode i3 (2^(i2-1))
		)
	lcnt = lcnt + num_faces_array[i2]
	)
	
	-- Set name and a few other parameters
	mnode.name = node_name
	mnode.parent = node_par
	mnode.wirecolor = color 224 198 87
	mnode.transform = node_M
	
	-- cull and unknown parameters
	node_cull = ReadByte fID #unsigned
	dummy = ReadShort fID #unsigned
	
	-- create command nodes "c_"
	-- position the command nodes here:
	cM = matrix3 [1, 0, 0] [0, 1, 0] [0, 0, 1] [12, 12, 0]
	
	-- Mesh blending mode
	thisMeshUsesAlpha = false
	thisMeshUsesAdditive = false
	if not(node_texBld == "default") then
	(
		Point transform:cM name:("c_tm_"+node_texBld) parent:mnode wirecolor:(color 42 7 132) size:10 axistripod:off cross:on -- dark purple
		if ((node_texBld == "alpha") or (node_texBld == "wormhole")) then
		(
			thisMeshUsesAlpha = true
		)
		else if (node_texBld == "additive") then
		(
			thisMeshUsesAdditive = true
		)
	)
	-- borg texture
	if not(node_borg == "") then
	(
		Point transform:cM name:("c_borg_"+node_borg) parent:mnode wirecolor:(color 42 7 132) size:10 axistripod:off cross:on -- dark purple
	)
	-- turn off backface cull (two-sided polies)
	if not(node_cull == 1) then
	(
		Point transform:cM name:("c_cullOff") parent:mnode wirecolor:(color 42 7 132) size:10 axistripod:off cross:on -- dark purple
	)
	-- custom lighting groups
	main_lm_name = group_lighting_mat[1]
		if (main_lm_name == undefined) then
	(
		main_lm = undefined
		main_lm_name = "undefined"
	)
	else
	(
		main_lm = meditmaterials[main_lm_name + "_01"]
	)

	-- check whether this material exists. Might be a standard material
	if (main_lm == undefined) then
	(
		-- material does not exist. Fill in basics
		num_lighting = num_lighting+1
		matTemp = standardmaterial shaderType:1
		matTemp.name = (main_lm_name + "_01")
		matTemp.diffuseMap = Bitmaptexture fileName:(node_texFile + ".tga") name:"Map #1" alphaSource:2
		matTemp.specularLevelMap = matTemp.diffuseMap
		if thisMeshUsesAlpha then
		(
			matTemp.opacityMap = matTemp.diffuseMap
			matTemp.opacityMap.alphaSource = 0
			matTemp.opacityMap.monoOutput = 1
		)
		if thisMeshUsesAdditive then
		(
			matTemp.opacityFallOff = 1
			matTemp.opacityType = 2
		)
		if not(node_bump == undefined) then
		(
			matTemp.bumpMap = Bitmaptexture fileName:(node_bump + ".tga") name:"Map #2"
		)			
		if (num_lighting < 25)  then
		(
			-- assign new material to the mesh
			meditMaterials[num_lighting] = copy matTemp
			mnode.material = meditMaterials[num_lighting]
			meditMaterials[num_lighting].ShowInViewport = on
		)
		else
		(
			matTemp.ShowInViewport = on
			mnode.material = copy matTemp
		)
	)
	else -- material already exists, this is the common scenario
	(
		if (main_lm.diffuseMap == undefined) then
		(
			-- material exits, but texture data is empty, so fill it
			main_lm.diffuseMap = Bitmaptexture fileName:(node_texFile + ".tga") name:"Map #1" alphaSource:2
			main_lm.specularLevelMap = main_lm.diffuseMap
			if thisMeshUsesAlpha then
			(
				main_lm.opacityMap = copy main_lm.diffuseMap
				main_lm.opacityMap.alphaSource = 0
				main_lm.opacityMap.monoOutput = 1
			)
			if thisMeshUsesAdditive then
			(
				main_lm.opacityFallOff = 1
				main_lm.opacityType = 2
			)
			if ((main_lm.selfillumMapEnable == on) or (not(main_lm.ambient == (color 0 0 0)))) then
			(
				main_lm.selfillumMap = copy main_lm.diffuseMap
				main_lm.selfillumMap.rgbOutput = 1
				main_lm.selfillumMapAmount = (main_lm.ambient.red + main_lm.ambient.green + main_lm.ambient.blue)/(3*2.55)
			)
			if not(node_bump == undefined) then
			(
				main_lm.bumpMap = Bitmaptexture fileName:(node_bump + ".tga") name:"Map #2"
			)
			-- assign material to the mesh
			mnode.material = meditmaterials[main_lm.name]
		)
		else 
		(
			-- find material that uses the same texture
			matchedMaterial = undefined
			for ii = 1 to (amin num_lighting 24) do
			(
				notOK = 4
				if (not(meditMaterials[ii].diffuseMap == undefined)) then
				(
					notOK = notOK - 1
					mapName = meditMaterials[ii].name
					-- remove trailing '_##' from material name.
					splitCmd = filterString mapName "_"
					mCount = splitCmd[splitCmd.count] as Float
					if (not(mCount == Undefined)) then
					(
						mapName = substring mapName 1 (mapName.count - splitCmd[splitCmd.count].count - 1)
					)
					-- see if all releant properties match
					if (mapName == main_lm_name) then
					(
						notOK = notOK - 1
					)
					if (not(node_texFile == undefined)) then
					(
						if (meditMaterials[ii].diffuseMap.filename == node_texFile + ".tga") then
						(
							notOK = notOK - 1
						)
					)
					if (meditMaterials[ii].opacityMapEnable == thisMeshUsesAlpha) then
					(
						notOK = notOK - 1
					)
					if ((not(meditMaterials[ii].bumpMap == undefined)) and (not(node_bump == undefined))) then
					(
						if (not(meditMaterials[ii].bumpMap.filename == node_bump + ".tga")) then
						(
							notOK = 1
						)
					)
				)
				if (notOK == 0) then
				(
					-- everything matches, so use this material
					matchedMaterial = meditMaterials[ii].name
					mnode.material = meditMaterials[ii]
					meditMaterials[ii].ShowInViewport = on
					exit true
				)
			)
			
			if (matchedMaterial == undefined) then
			(
				-- material already exists with texture, so make a copy
				num_lighting = num_lighting+1
				matTemp = standardMaterial shaderType:1
				matTemp = copy meditMaterials[main_lm.name]
				copy_lm = matTemp
				
				-- increment the material name
				mID = main_lm.name
				mIDc = mID.count
				nameSpl = filterString mID "_"
				--cnt = (nameSpl[nameSpl.count] as number) + 1
				cnt = num_lighting
				if (cnt < 10) then (cntStr = "0"+(cnt as string))
				else (cntStr = cnt as string)
				mID[mIDc] = cntStr[2]
				mID[mIDc-1] = cntStr[1]
				copy_lm.name = mID
				
				-- update the texture maps
				copy_lm.diffuseMap = Bitmaptexture fileName:(node_texFile + ".tga") name:"Map #1" alphaSource:2
				copy_lm.specularLevelMap = copy_lm.diffuseMap
				if thisMeshUsesAlpha then
				(
					copy_lm.opacityMap = copy copy_lm.diffuseMap
					copy_lm.opacityMap.alphaSource = 0
					copy_lm.opacityMap.monoOutput = 1
				)
				else
				(
					copy_lm.opacityMap = undefined
				)
				if thisMeshUsesAdditive then
				(
					copy_lm.opacityFallOff = 1
					copy_lm.opacityType = 2
				)
				if ((copy_lm.selfillumMapEnable == on) or (not(main_lm.ambient == (color 0 0 0)))) then
				(
					copy_lm.selfillumMap = copy copy_lm.diffuseMap
					copy_lm.selfillumMap.rgbOutput = 1
					copy_lm.selfillumMapAmount = (copy_lm.ambient.red + copy_lm.ambient.green + copy_lm.ambient.blue)/(3*2.55)
				)
				if not(node_bump == undefined) then
				(
					copy_lm.bumpMap = Bitmaptexture fileName:(node_bump + ".tga") name:"Map #2"
				)
				else
				(
					copy_lm.bumpMap = undefined
				)
				-- assign copied material to the mesh
				mnode.material = matTemp
				matTemp.ShowInViewport = on
				
				if (num_lighting<25) then
				(
					meditMaterials[num_lighting] = copy matTemp
					mnode.material = meditMaterials[num_lighting]
					meditMaterials[num_lighting].ShowInViewport = on
				)
			)
		) 
	)-- end main lighting material
	
	-- fill the remaining lighting groups via c_lm nodes
	if (node_lgroups > 1) then
	(
		multMat = multiMaterial numsubs:node_lgroups
		multMat[1] = mnode.material
		mnode.material = multMat
	)
	p = convertToPoly(mnode)
	for i1 = 2 to group_lighting_mat.count do
	(
		nameStr = "c_lm_" + (i1 as string) + "_" + group_lighting_mat[i1]
		Point transform:cM name:nameStr parent:mnode wirecolor:(color 42 7 132) size:10 axistripod:off cross:on -- dark purple
		
		-- part of the mesh must use a different material. Find this material and assign to the smoothing group
		newMat = copy mnode.material[1]
		newMat.ShowInViewport = on
			
		for ii = 1 to (amin num_lighting 24) do
		(
			-- remove trailing '_##' from material name.
			mapName = meditMaterials[ii].name
			splitCmd = filterString mapName "_"
			mCount = splitCmd[splitCmd.count] as Float
			if (not(mCount == Undefined)) then
			(
				mapName = substring mapName 1 (mapName.count - splitCmd[splitCmd.count].count - 1)
			)
			-- see if name matches
			if (mapName == group_lighting_mat[i1]) then
			(
				refMat = meditMaterials[ii]
				newMat.ambient = refMat.ambient
				newMat.diffuse = refMat.diffuse
				newMat.specular = refMat.specular
				newMat.specularLevel = refMat.specularLevel
				newMat.glossiness = refMat.glossiness
			)
		)
		multMat[i1] = newMat
		
		-- select this smoothing group
		p.selectBySmoothGroup i1
		faces = polyOp.getFaceSelection p -- this is a bit array. 1: selected, 0: deselected. Length = numfaces
		polyop.setFaceMatID p faces i1
	)
	mnode = ConvertToMesh p -- convert back to mesh
	
) -- end node type:"mesh"

) -- end cycle through all nodes

) -- end WriteNodes

-- creates animation keyframes from the SOD file
function ReadAnimTransforms fID =
(
include "SODImportSupport.ms"

sliderTime = 0f
num_anim_obj = ReadShort fID #unsigned
cM = matrix3 [1, 0, 0] [0, 1, 0] [0, 0, 1] [8, 8, 0]
maxFrame = 0

for i1 = 1 to num_anim_obj do
(
	-- Basic animation chanel parameters
	obj_name = ReadIdentifier fID
	obj = nodeByName obj_name
	print obj_name
	num_frames = ReadShort fID #unsigned
	anim_len = ReadFloat fID
	anim_type = ReadShort fID #unsigned
	maxFrame = amax num_frames maxFrame	
	-- create c_keys_ node
	nameStr = "c_keys_" + (anim_len as string) + "_" + (num_frames as string)
	Point transform:cM name:nameStr parent:obj wirecolor:(color 215 125 50) size:10 axistripod:off cross:on
	
	-- Read in the transforms
	if (anim_type == 0) then
	(
		print anim_type
		with animate on
		(
			for i2 = 1 to num_frames do
			(
				-- The local parent-to-child transform is the anim. controller "value"
				M = XYZtoRUF (ReadTransform fID)
				
				-- Create the keyframe
				at time (i2-1) obj.rotation.controller.value = M.rotation
				at time (i2-1) obj.position.controller.value = M.position
			)
		)
	)-- end read pose transforms
	else if (anim_type == 5) then
	(
		print anim_type
		with animate on
		(
			for i2 = 1 to num_frames do
			(
				-- Animation scale key frames
				S = ReadFloat fID
				at time (i2-1) obj.scale.controller.value = point3 S S S
			) 
		)
	) -- end read scale animation
	
	-- simplify the resulting keys
	-- note: interval leaves the last frame intact
	reduceKeys obj.transform.controller 0.5 1 (interval 0 (num_frames-2))
	animationRange = interval 0 maxFrame
		
) -- end read animation channels

) -- end ReadAnimTransforms

function ReadAnimTexRefs fID =
(
	include "SODImportSupport.ms"

	cM = matrix3 [1, 0, 0] [0, 1, 0] [0, 0, 1] [12, 8, 0]
	num_anim = ReadShort fID #unsigned
	for i1 = 1 to num_anim do
	(
		dummy = ReadByte fID #unsigned
		obj = nodeByName (ReadIdentifier fID)
		animID = ReadIdentifier fID
		animOffset = ReadFloat fID
		nameStr = ""
		if (animOffset == 0) then
		(
			nameStr = "c_anim_" + animID
		)
		else
		(
			nameStr = "c_anim_" + animID + "_" + animOffset
		)
		Point transform:cM name:nameStr parent:obj wirecolor:(color 215 125 50) size:10
	)
) -- end ReadAnimTexRefs
fn HideNonRenderNodes =
(
	objects.isHidden = true
	geometry.isHidden = false
	--showNodeTree 
)
-- MAIN
	-- get file for opening
	if (FullFilePath.count == 0) then
	(
		try fID = GetSaveFile() 
		catch
		(
		messageBox "Error opening file for importing"
		fclose fID
		fID = Undefined
		)
	)
	else
	(
		try (fID = fopen FullFilePath "rb")
		catch
		(
		messageBox "Error opening the specified file"
		fID = undefined
		)
	)
	if not (fID == Undefined) then
	(
	try
	(
		disableSceneRedraw ()
		resetEnvironment ()
		ReadHeader fID
		ReadLightingMaterials fID
		ReadNodes fID TexturePath
		ReadAnimTransforms fID
		ReadAnimTexRefs fID
		HideNonRenderNodes ()
		print "SOD Imported"
		messageBox "Use responsibly!  Modding communities are built on trust and respect for fellow modders. Please respect the work of those that came before you."
	)
	catch 
	(
		messageBox "Import Error!"
		--format "MESSAGE: % \n\n %" (getCurrentException()) (stack())
		--stack ()
	)
	)
enableSceneRedraw()

fclose fID
-- end import	
)-- end MAIN

-- Auto-run the script
ImportSODRender ()